
local global_env = _G

--- Utility functions shared by other modules.
module("luarocks.util", package.seeall)

local config = require("luarocks.config")

local scheduled_functions = {}

--- Schedule a function to be executed upon program termination.
-- This is useful for actions such as deleting temporary directories
-- or failure rollbacks.
-- @param f function: Function to be executed.
-- @param ... arguments to be passed to function.
-- @return table: A token representing the scheduled execution,
-- which can be used to remove the item later from the list.
function schedule_function(f, ...)
   assert(type(f) == "function")
   
   local item = { fn = f, args = {...} }
   table.insert(scheduled_functions, item)
   return item
end

--- Unschedule a function.
-- This is useful for cancelling a rollback of a completed operation.
-- @param table: The token representing the scheduled function that was
-- returned from the schedule_function call.
function remove_scheduled_function(item)
   for k, v in pairs(scheduled_functions) do
      if v == item then
         table.remove(scheduled_functions, k)
         return
      end
   end
end

--- Execute scheduled functions.
-- Some calls create temporary files and/or directories and register
-- corresponding cleanup functions. Calling this function will run
-- these function, erasing temporaries.
-- Functions are executed in the inverse order they were scheduled.
function run_scheduled_functions()
   for i = #scheduled_functions, 1, -1 do
      local item = scheduled_functions[i]
      item.fn(unpack(item.args))
   end
end

--- Implements recursion for the platform_overrides function
-- @see platform_overrides   
local function deep_replace(tbl, platform_tbl)
   for k, v in pairs(platform_tbl) do
      if type(v) == "table" then
         if not tbl[k] then
            tbl[k] = {}
         end
         deep_replace(tbl[k], v)
      else
         tbl[k] = v
      end
   end
end

--- Perform platform-specific overrides on a table.
-- Overrides values of table with the contents of the appropriate
-- subset of its "platforms" field. The "platforms" field should
-- be a table containing subtables keyed with strings representing
-- platform names. Names that match the contents of the global
-- config.platforms setting are used. For example, if
-- config.platforms= {"plat"}, then the fields of
-- tbl.platforms.plat will overwrite those of tbl with the same
-- names. For table values, the operation is performed recursively
-- (tbl.platforms.plat.x.y.z overrides tbl.x.y.z; other contents of
-- tbl.x are preserved).
-- @param tbl table: Table which may contain a "platforms" field;
-- if it doesn't, this function does nothing.
function platform_overrides(tbl)
   assert(type(tbl) == "table")
   
   if tbl.platforms then
      for _, platform in ipairs(config.platforms) do
         local platform_tbl = tbl.platforms[platform]
         if platform_tbl then
            deep_replace(tbl, platform_tbl)
         end
      end
   end
end

--- Perform make-style variable substitutions on string values of a table.
-- For every string value tbl.x which contains a substring of the format
-- "$(XYZ)" will have this substring replaced by vars["XYZ"], if that field
-- exists in vars. Only string values are processed; this function
-- does not scan subtables recursively.
-- @param tbl table: Table to have its string values modified.
-- @param vars table: Table containing string-string key-value pairs 
-- representing variables to replace in the strings values of tbl.
function variable_substitutions(tbl, vars)
   assert(type(tbl) == "table")
   assert(type(vars) == "table")
   
   local updated = {}
   for k, v in pairs(tbl) do
      if type(v) == "string" then
         updated[k] = v:gsub("%$%(([%a_]+)%)", vars)
      end
   end
   for k, v in pairs(updated) do
      tbl[k] = v
   end
end
